import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'dart:typed_data';
import 'crypto.dart';
import 'package:http/http.dart' as http;
import 'net_addr.dart';
import 'package:messagepack/messagepack.dart' as msgpack;
class Tail {
  RawDatagramSocket _socket;
  final NetAddr host;

  static Future<Tail> connect(NetAddr host) async {
    return Tail(await RawDatagramSocket.bind("0.0.0.0", 0), host);
  }

  Tail(this._socket, this.host) {
    _reload();
    Timer.periodic(Duration(seconds: 10), (timer) {
      _reload();
    });
  }

  Future<void> checkUDP(List<int> last) async {
    try {
      _socket.send(last, (await host.lookup()).first, host.port);
    } catch (_) {}
  }

  Stream<Uint8List> get stream {
    return _stream.stream;
  }

  void close() {
    _socket.close();
    _cli.close();
  }

  final _stream = StreamController<Uint8List>();
  bool _thereIsAnActiveHTTPClient = false;

  Future<void> _reload() async {
    try {
      _cli.close();
      _socket.close();
    } catch (_) {
      print(_);
    }
    _cli = http.Client();
    _socket = await RawDatagramSocket.bind("0.0.0.0", 0);
    _socket
        .asyncMap((event) {
          return _socket.receive()?.data;
        })
        .where((event) => event != null)
        .asyncMap((event) => event!)
        .forEach((element) {
          _stream.add(element);
        });
  }

  Future<void> checkHTTP(List<int> last) async {
    if (_thereIsAnActiveHTTPClient) {
      return;
    }
    _thereIsAnActiveHTTPClient = true;
    try {
      await _checkHTTP(last);
    } finally {
      _thereIsAnActiveHTTPClient = false;
    }
  }

  http.Client _cli = http.Client();

  Future<void> _checkHTTP(List<int> last) async {
    try {
      if (_stream.isPaused || _stream.isClosed) {
        return;
      }
      var resp = await _cli.put(
          host.toUri(scheme: "http", queryParameters: {"limit": "10"}),
          body: last);
      var pack = msgpack.Unpacker(resp.bodyBytes);
      var len = pack.unpackListLength();
      for (var i=0;i<len;i++){
        _stream.add(Uint8List.fromList(pack.unpackBinary()));
      }
    } catch (_) {}
  }

  static Future<void> push(
      NetAddr addr, List<int> data, List<int> sk, Duration timeout) async {
    var cli = http.Client();
    var timer = Timer(timeout, () {
      cli.close();
    });
    try {
      var resp = await cli.post(addr.toUri(scheme: "http"),
          headers: {
            "Content-Signature": base64UrlEncode(await Crypto.sign(sk, data))
          },
          body: data);
      if (resp.statusCode != 204) {
        throw Exception(resp.body.isEmpty ? resp.statusCode : resp.body);
      }
    } finally {
      cli.close();
      if (timer.isActive) {
        timer.cancel();
      }
    }
  }
}
